<?php defined("CYCLONE") or die();

/**
 * Cyclone - Cylindrical projection tool for photogrammetry.
 * Copyright (C) 2015  Bernhard Arnold <bernhard.arnold@burgried.at>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

require_once "system/classes/points.class.php";

define('FORMAT_WKT', 'wkt');
define('FORMAT_CSV', 'csv');
define('FORMAT_JSON', 'json');
define('EOL', "\r\n"); // take care for windows people.

function post($name, $default = '') {
    return isset($_POST[$name]) ? htmlspecialchars($_POST[$name]) : $default;
}

function toWKT($points) {
    $result = array();
    foreach ($points as &$point) {
        $result[] = sprintf("Point ( %.3f  %.3f )", $point->x(), $point->y());
    }
    return implode($result, EOL);
}

function toCSV($points) {
    $result = array("Name; X; Y; Code");
    foreach ($points as &$point) {
        $result[] = sprintf("%s; %.3f; %.3f; %s", $point->name(), $point->x(), $point->y(), $point->code());
    }
    return implode($result, EOL);
}

function toJSON($points) {
    $result = array("{".EOL."  [");
    foreach ($points as &$point) {
        $coord = array("name" => $point->name(), "x" => $point->x(), "y" => $point->y(), "code" => $point->code());
        $result[] = "    ".json_encode($coord).($i + 1 < sizeof($points) ? ',' : '');
    }
    $result[] = "  ]".EOL."}";
    return implode($result, EOL);
}

function autoFormat($format, $points) {
    switch ($format) {
        default:
        case FORMAT_WKT:
            return toWKT($points);
        case FORMAT_CSV:
            return toCSV($points);
        case FORMAT_JSON:
            return toJSON($points);
    }
}

class Cyclone {
    private function load_view($name, $attr = array()) {
        extract($attr);
        include "system/views/$name.inc.php";
    }

    private function read_points($str, &$errors) {
        $points = array();
        $line_nr = 1;
        $lines = preg_split('/[\r\n]+/', $str);
        foreach ($lines as $line) {
            $line = trim($line);
            if ($line) {
                try {
                    $points[] = ArchServPoint::fromString($line);
                } catch (Exception $e) {
                    $errors[] = "line ".$line_nr.": ".$e->getMessage();
                }
            }
        }
        return $points;
    }

    private function _sector_angle($m, $p, $a) {
        // Calculates sector offsets and triangle placement.
        $s = pi() / 2.;
        if ($p->x() >= $m->x() and $p->y() >= $m->y())
            return $a + $s * 0.;
        if ($p->x() < $m->x() and $p->y() >= $m->y())
            return ($s - $a) + $s * 1.;
        if ($p->x() < $m->x() and $p->y() < $m->y())
            return $a + $s * 2.;
        if ($p->x() >= $m->x() and $p->y() < $m->y())
            return ($s - $a) + $s * 3.;
    }

    private function _angle($m, $p) {
        // Returns angle a* for triangle with hypotenuse c* between m* and p*.
        $d = $m->deltaXY($p);
        $c = sqrt( pow($d[0], 2) + pow($d[1], 2) );
        if ($c == 0)
            return 0.0;
        $a = asin( $d[1] / $c );
        return $this->_sector_angle($m, $p, $a);
    }

    private function transform($r, $m, $p) {
        return $this->_angle($m, $p) * $r;
    }

    private function calcOffset($r, $m, $p) {
        return $this->transform($r, $m, $p);
    }

    public function exec_() {
        $errors = array();

        //
        // Initial values.
        //
        $center = new Point2D(0, 0);
        $radius = 1.0;
        $start = new Point2D(0, 0);
        $points = array();
        $format = FORMAT_WKT;

        //
        // Get user input.
        //
        if (post('calculate')) {
            try {
                $center = Point2D::fromString(trim(post('center')));
            } catch (Exception $e) {
                $errors[] = "center point: ".$e->getMessage();
            }
            $radius = floatval(trim(post('radius')));
            try {
                $start = Point2D::fromString(trim(post('start')));
            } catch (Exception $e) {
                $errors[] = "start point: ".$e->getMessage();
            }
            $points = $this->read_points(post('points'), $errors);
            $format = post('format');
        }

        //
        // List input points
        //

        // Calculate the extent.
        $extent = 2 * pi() * $radius;

        // Calculcate offset for first point (most left point).
        $offset = $this->calcOffset($radius, $center, $start);

        $result_points = array();

        foreach ($points as $i => &$point) {
            // Transform point
            $x = $this->transform($radius, $center, $point);
            // Apply offset for startingpoint other then 0 [rad].
            if ($x < $offset) {
                $x += $extent;
            }
            $x = $x - $offset;
            // Create new transformed point in ArchServ format (to preserve name and code).
            $result_points[] = new ArchServPoint($point->name(), $x, $point->z(), 0.0, $point->code());
        }

        $result_ascii = autoFormat($format, $result_points);

        $this->load_view('document', array(
            'errors' => sizeof($errors) ? "<p class=\"error\">".implode($errors, "</p><p class=\"error\">")."</p>" : "",
            'center' => post('center', '0, 0'),
            'radius' => post('radius', '10'),
            'start' => post('start', '0, 0'),
            'points' => post('points'),
            'format' => $format,
            'c_center' => sprintf("%s", $center),
            'c_radius' => sprintf("%.3f", $radius),
            'c_extent' => sprintf("%.3f", $extent),
            'c_start' => sprintf("%s", $start),
            'c_offset' => sprintf("%.3f", $offset),
            'result' => $result_ascii
        ));
    }

} // class Mercator
